import { isArray } from "@thi.ng/checks/is-array";
import { isPlainObject } from "@thi.ng/checks/is-plain-object";
import { isString } from "@thi.ng/checks/is-string";
import { split } from "@thi.ng/strings/split";
import { wordWrapLine, wordWrapLines } from "@thi.ng/strings/word-wrap";
/**
 * Returns true iff `x` is a {@link WasmPrim32}.
 *
 * @param x
 */
export const isNumeric = (x) => /^(([iu](8|16|32))|(f(32|64)))$/.test(x);
/**
 * Returns true iff `x` is a `i64` or `u64`.
 *
 * @param x
 */
export const isBigNumeric = (x) => /^[iu]64$/.test(x);
export const isSizeT = (x) => /^[iu]size$/.test(x);
/**
 * Returns true iff `x` is a {@link WasmPrim}.
 *
 * @param x
 */
export const isWasmPrim = (x) => isNumeric(x) || isBigNumeric(x);
export const isWasmString = (x) => x === "string";
export const isPadding = (f) => f.pad != null && f.pad > 0;
export const isPointer = (x) => x === "ptr";
export const isFuncPointer = (type, coll) => coll[type]?.type === "funcptr";
export const isEnum = (type, coll) => coll[type]?.type === "enum";
export const isSlice = (x) => x === "slice";
export const isOpaque = (x) => x === "opaque";
/**
 * Returns true iff the struct field is a pointer, slice or "string" type
 *
 * @param f
 */
export const isPointerLike = (f, coll) => isPointer(f.tag) ||
    isSlice(f.tag) ||
    isWasmString(f.type) ||
    isOpaque(f.type) ||
    isFuncPointer(f.type, coll);
/**
 * Returns true if `type` is "slice".
 *
 * @param type
 */
export const isStringSlice = (type) => type === "slice";
export const isStruct = (x) => x.type === "struct";
export const isUnion = (x) => x.type === "union";
/**
 * Returns filtered array of struct fields of with "ptr" tag or function
 * pointers.
 *
 * @param fields
 *
 * @internal
 */
export const pointerFields = (fields) => fields.filter((f) => isPointer(f.tag));
/**
 * Returns filtered array of struct fields of only "string" fields.
 *
 * @param fields
 *
 * @internal
 */
export const stringFields = (fields) => fields.filter((f) => isWasmString(f.type) && f.tag !== "ptr");
export const sliceTypes = (coll) => new Set(Object.values(coll)
    .flatMap((x) => (isStruct(x) || isUnion(x) ? x.fields : []))
    .map((x) => (x.tag === "slice" ? x.type : null))
    .filter((x) => !!x));
/**
 * Returns enum identifier formatted according to given opts.
 *
 * @param opts
 * @param name
 *
 * @internal
 */
export const enumName = (opts, name) => opts.uppercaseEnums ? name.toUpperCase() : name;
/**
 * Returns given field's default value (or undefined). The `lang` ID is required
 * to obtain the language specific value if the default is given as object.
 *
 * @param f
 * @param lang
 */
export const defaultValue = (f, lang) => f.default !== undefined
    ? isPlainObject(f.default)
        ? f.default[lang]
        : f.default
    : undefined;
/**
 * Takes an array of strings or splits given string into lines, word wraps and
 * then prefixes each line with given `width` and `prefix`. Returns array of new
 * lines.
 *
 * @param prefix
 * @param str
 * @param width
 */
export const prefixLines = (prefix, str, width) => (isString(str)
    ? wordWrapLines(str, { width: width - prefix.length })
    : str.flatMap((x) => wordWrapLine(x, { width: width - prefix.length }))).map((line) => prefix + line);
export const ensureLines = (src, key) => isString(src)
    ? split(src)
    : isArray(src)
        ? src
        : key
            ? src[key]
                ? ensureLines(src[key], key)
                : []
            : [];
export const ensureStringArray = (src) => isString(src) ? [src] : src;
/**
 * Yields iterator of given lines, each with applied indentation based on given
 * scope regexp's which are applied to each line to increase or decrease
 * indentation level (the initial indentation level can be specified via
 * optional `level` arg, default 0). If `scopeStart` succeeds, the indent is
 * increased for the _next_ line. If `scopeEnd` succeeds the level is decreased
 * for the _current_ line. ...
 *
 * @param lines
 * @param indent
 * @param scopeStart
 * @param scopeEnd
 * @param level
 */
export function* withIndentation(lines, indent, scopeStart, scopeEnd, level = 0) {
    const stack = new Array(level).fill(indent);
    for (let l of lines) {
        scopeEnd.test(l) && stack.pop();
        const curr = stack.length ? stack[stack.length - 1] : "";
        yield curr + l;
        scopeStart.test(l) && stack.push(curr + indent);
    }
}
